==> dovecot-0.91.tar.gz: dovecot-0.91/src/lib-storage/index/index-save.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "iobuffer.h"
#include "index-storage.h"

#include <stdlib.h>
#include <unistd.h>

static int write_with_crlf(int fd, const unsigned char *data, int size,
			   int *last_cr)
{
	int cr, i;

	cr = *last_cr ? -1 : -2;
	for (i = 0; i < size; i++) {
		if (data[i] == '\r')
			cr = i;
		else if (data[i] == '\n' && cr != i-1) {
			/* missing CR */
			if (write(fd, data, i) != i)
				return FALSE;
			if (write(fd, "\r", 1) != 1)
				return FALSE;

			/* skip the data so far. \n is left into buffer and
			   we'll continue from the next character. */
			data += i;
			size -= i;
			i = 0; cr = -2;
		}
	}

	return write(fd, data, size) == i;
}

int index_storage_save_into_fd(MailStorage *storage, int fd, const char *path,
			       IOBuffer *buf, off_t data_size)
{
	unsigned char *data;
	int ret, size, set_insert_crlf, failed, last_cr;

	set_insert_crlf = getenv("SAVE_WITH_CRLF") != NULL;
	last_cr = FALSE;

	while ((ret = io_buffer_read(buf)) != 0) {
		if (ret == -1) {
			mail_storage_set_critical(storage, "Error reading "
						  "mail: %s", strerror(errno));
			return FALSE;
		}

		/* -2 = buffer full, ignore it since we're just emptying it.. */

		data = io_buffer_get_data(buf, &size);
		if (size == 0)
			continue;

		if (size > data_size)
			size = data_size;
		buf->pos += size;
		data_size -= size;

		failed = set_insert_crlf ?
			!write_with_crlf(fd, data, size, &last_cr) :
			write(fd, data, size) != size;

		if (failed) {
			mail_storage_set_critical(storage, "write() failed "
						  "for file %s: %s",
						  path, strerror(errno));
			return FALSE;
		}
	}

	return TRUE;
}

==> dovecot-0.95.tar.gz: dovecot-0.95/src/lib-storage/index/index-save.c <==
/*
    Copyright (C) 2002 Timo Sirainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "lib.h"
#include "iobuffer.h"
#include "index-storage.h"

#include <stdlib.h>
#include <unistd.h>

static int write_with_crlf(int fd, const unsigned char *data, int size,
			   int *last_cr)
{
	int cr, i;

	cr = *last_cr ? -1 : -2;
	for (i = 0; i < size; i++) {
		if (data[i] == '\r')
			cr = i;
		else if (data[i] == '\n' && cr != i-1) {
			/* missing CR */
			if (write(fd, data, i) != i)
				return FALSE;
			if (write(fd, "\r", 1) != 1)
				return FALSE;

			/* skip the data so far. \n is left into buffer and
			   we'll continue from the next character. */
			data += i;
			size -= i;
			i = 0; cr = -2;
		}
	}

	return write(fd, data, size) == i;
}

int index_storage_save_into_fd(MailStorage *storage, int fd, const char *path,
			       IOBuffer *buf, off_t data_size)
{
	unsigned char *data;
	int ret, size, last_cr;

	last_cr = FALSE;

	while ((ret = io_buffer_read(buf)) != 0) {
		if (ret == -1) {
			mail_storage_set_critical(storage, "Error reading "
						  "mail: %s", strerror(errno));
			return FALSE;
		}

		/* -2 = buffer full, ignore it since we're just emptying it.. */

		data = io_buffer_get_data(buf, &size);
		if (size == 0)
			continue;

		if (size > data_size)
			size = data_size;
		buf->pos += size;
		data_size -= size;

		if (write_with_crlf(fd, data, size, &last_cr)) {
			mail_storage_set_critical(storage, "write() failed "
						  "for file %s: %s",
						  path, strerror(errno));
			return FALSE;
		}
	}

	return TRUE;
}
